/*
  Copyright (C) 2003 Nicol N Schraudolph <nic@schraudolph.org>

  This is free software with ABSOLUTELY NO WARRANTY.

  This library is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2.1 of the License, or
  (at your option) any later version.

  This library is distributed in the hope that it will be useful, but
  WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this library; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
  USA
*/

// NB: SIMPLIFIED VERSION FOR CRF++ CODE
// forward automatic differentiation class template
// T = float, double, long double, complex<>, Fad<>
//
// (c) 2003 Nicol N. Schraudolph, all rights reserved

#ifndef FAD_H
#define FAD_H

#include <cmath>
#include <vector>
#include <iostream>

#define CLASS Fad<T>
#define TEMPL template <class T>

template <class T = double> class Fad
{
private:
    T value;
    T deriv;

public:
    // constructors:
    //
    inline CLASS(const T v = T(0.0)) : value(v), deriv(T(0.0)) {};
    inline CLASS(const CLASS& x) : value(x.v()), deriv(x.d()) {};

    // assignment:
    //
    inline CLASS& operator=(const CLASS& x)
    {
        value = x.v();
        deriv = x.d();
        return *this;
    }
    inline CLASS& operator=(const T& x)
    {
        value = x;
        deriv = T(0.0);
        return *this;
    }

    // accessors:
    //
    inline       T&         v(void)       { return value; }
    inline const T&         v(void) const { return value; }
    inline       T& operator*(void)       { return value; }
    inline const T& operator*(void) const { return value; }

    inline       T&         d(void)       { return deriv; }
    inline const T&         d(void) const { return deriv; }

    // I/O format: value[deriv]
    //
    inline friend std::ostream& operator<< (std::ostream& s, const CLASS& x)
    {
        return s << x.v() << '[' << x.d() << ']';
    }
    inline friend std::istream& operator>> (std::istream& s, CLASS& x)
    {
        return s >> x.v();   // NEEDS WORK
    }

    inline CLASS& operator++(void)
    {
        ++value;
        return *this;
    }
    inline CLASS& operator--(void)
    {
        --value;
        return *this;
    }

    inline CLASS& operator+=(const T& x)
    {
        value += x;
        return *this;
    }
    inline CLASS& operator-=(const T& x)
    {
        value -= x;
        return *this;
    }
    inline CLASS& operator*=(const T& x)
    {
        value *= x;
        deriv *= x;
        return *this;
    }
    inline CLASS& operator/=(const T& x)
    {
        value /= x;
        deriv /= x;
        return *this;
    }

    inline CLASS& operator+=(const CLASS& x)
    {
        value += x.v();
        deriv += x.d();
        return *this;
    }
    inline CLASS& operator-=(const CLASS& x)
    {
        value -= x.v();
        deriv -= x.d();
        return *this;
    }
    inline CLASS& operator*=(const CLASS& x)
    {
        value *= x.v();
        deriv *= x.v();
        deriv += value*x.d();
        return *this;
    }
    inline CLASS& operator/=(const CLASS& x)
    {
        value /= x.v();
        deriv /= x.v();
        deriv -= x.d()*value/x.v();
        return *this;
    }

};  // CLASS


// load differentiation rules:
//
#define REG(G) T tmp(G); r.d() = tmp*x.d()
#define REG2(G1,G2) T t1(G1); T t2(G2); r.d() = t1*a.d() + t2*b.d()
#include "rules.h"

#endif  // FAD_H

