
// automatic differentiation rules macro file
//
// (c) 2003 Nicol N. Schraudolph, all rights reserved

// sanity checks:
//
#ifndef FAD_H
#ifndef RAD_H
#error never include this file directly, use fad.h or rad.h!
#endif
#endif

#ifndef TEMPL
#error TEMPL undefined, include fad.h or rad.h!
#endif
#ifndef CLASS
#error CLASS undefined, include fad.h or rad.h!
#endif
#ifndef REG
#error REG undefined, include fad.h or rad.h!
#endif
#ifndef REG2
#error REG2 undefined, include fad.h or rad.h!
#endif


// initialize static members:
//
// TEMPL int CLASS::num = 0;
// TEMPL CLASS* CLASS::marks[N];


// assignment operators:
//
#define ASSOPS(OP) TEMPL inline CLASS operator \
OP##=(CLASS& r, const T& x) {return r = r OP x;} TEMPL inline \
CLASS operator OP##=(CLASS& r, const CLASS& x) {return r = r OP x;}

// ASSOPS(+)
// ASSOPS(-)
#ifdef RAD_H
ASSOPS(*)
ASSOPS(/)
#endif

// binary operators:
//
#define BINOP0(OP,G1,G2) \
TEMPL inline CLASS operator OP(const CLASS& a, const CLASS& b) { \
CLASS r((*a) OP (*b)); REG2(G1,G2); return r; }

// only first argument active:
//
#define BINOP1(OP,G) \
TEMPL inline CLASS operator OP(const CLASS& x, const T& b) { \
CLASS r((*x) OP b); REG(G); return r; }

// only second argument active:
//
#define BINOP2(OP,G) \
TEMPL inline CLASS operator OP(const T& a, const CLASS& x) { \
CLASS r(a OP (*x)); REG(G); return r; }

BINOP0(+, T(1), T(1))
BINOP1(+, T(1))
BINOP2(+, T(1))

BINOP0(-, T(1), T(-1))
BINOP1(-, T(1))
BINOP2(-, T(-1))

BINOP0(*, (*b), (*a))
BINOP1(*, b)
BINOP2(*, a)

BINOP0(/, (*b)/((*b)*(*b)), -(*a)/((*b)*(*b)))
BINOP1(/, T(1)/b)
BINOP2(/, -a/((*x)*(*x)))

/*** not worth it
#define PINOP(OP) \
TEMPL inline CLASS operator OP(const CLASS& a, const CLASS& b) \
{ CLASS r(a); r OP##= b; return r; } \
TEMPL inline CLASS operator OP(const CLASS& a, const T& b) \
{ CLASS r(a); r OP##= b; return r; } \
TEMPL inline CLASS operator OP(const T& a, const CLASS& b) \
{ CLASS r(b); r OP##= a; return r; }
***/


// binary functions:
//
#define BINARY0(FUNC,G1,G2) \
TEMPL inline CLASS FUNC(const CLASS& a, const CLASS& b) { \
CLASS r(FUNC(*a, *b)); REG2(G1,G2); return r; }

// only first argument active:
//
#define BINARY1(FUNC,G) \
TEMPL inline CLASS FUNC(const CLASS& x, const T& b) { \
CLASS r(FUNC(*x, b)); REG(G); return r; }

// only second argument active:
//
#define BINARY2(FUNC,G) \
TEMPL inline CLASS FUNC(const T& a, const CLASS& x) { \
CLASS r(FUNC(a, *x)); REG(G); return r; }

BINARY0(pow, (*r)*(*b)/(*a), (*r)*log(*a))
BINARY1(pow, (*r)*b/(*x))
BINARY2(pow, (*r)*log(a))

BINARY0(hypot, (*a)/(*r), (*b)/(*r))
BINARY1(hypot, (*x)/(*r))
BINARY2(hypot, (*x)/(*r))

BINARY0(atan2,  (*b)/((*a)*(*a) + (*b)*(*b)), -(*a)/((*a)*(*a) + (*b)*(*b)))
BINARY1(atan2,  b/((*x)*(*x) + b*b))
BINARY2(atan2, -a/(a*a + (*x)*(*x)))

BINARY0(copysign, copysign(T(1), (*a)*(*b)), T(0))
BINARY1(copysign, copysign(T(1), (*x)*b))
BINARY2(copysign, T(0))


// value comparison operators (passive result):
//
#define COMPOPS(OP) TEMPL inline bool \
operator OP(const CLASS& a, const CLASS& b) {return (*a) OP (*b);} \
TEMPL inline bool operator OP(const CLASS& x, const T& b) {return (*x) OP b;} \
TEMPL inline bool operator OP(const T& a, const CLASS& x) {return a OP (*x);}

COMPOPS(==)
COMPOPS(!=)
COMPOPS(>=)
COMPOPS(<=)
COMPOPS(>)
COMPOPS(<)


// unary operators:
//
#define UNOP(OP,GRAD) \
TEMPL inline CLASS operator OP(const CLASS& x) { \
CLASS r(OP *x); REG(GRAD); return r; }

UNOP(+, T(1))
UNOP(-, T(-1))


// unary functions:
//
#define UNARY(FUNC,GRAD) \
TEMPL inline CLASS FUNC(const CLASS& x) { \
CLASS r(FUNC(*x)); REG(GRAD); return r; }

UNARY(fabs, copysign(T(1), *x))

// exp/log/powers

UNARY(exp,   (*r))
UNARY(expm1, (*r) + T(1))

#define INVLN10 0.4342944819032518276454794131646864L
UNARY(log,   T(1)/(*x))
UNARY(log10, T(INVLN10)/(*x))
UNARY(log1p, T(1)/((*x) + T(1)))

UNARY(sqrt, T(1)/((*r) + (*r)))
UNARY(cbrt, T(1)/(T(3)*(*r)*(*r)))

// trigonometry

UNARY(sin,  cos(*x))
UNARY(cos, -sin(*x))
UNARY(tan,  T(1) + (*r)*(*r))

UNARY(sinh, cosh(*x))
UNARY(cosh, sinh(*x))
UNARY(tanh, T(1) - (*r)*(*r))

UNARY(asin,  T(1)/sqrt(T(1) - (*x)*(*x)))
UNARY(acos, T(-1)/sqrt(T(1) - (*x)*(*x)))
UNARY(atan,  T(1)/(T(1) + (*x)*(*x)))

UNARY(asinh, T(1)/sqrt((*x)*(*x) + T(1)))
UNARY(acosh, T(1)/sqrt((*x)*(*x) - T(1)))
UNARY(atanh, T(1)/(T(1) - (*x)*(*x)))

// special functions

#define TWORTPI 1.1283791670955125738961589031215452L  // 2/sqrt(pi)
UNARY(erf,   T(TWORTPI)*exp(-(*x)*(*x)))
UNARY(erfc, T(-TWORTPI)*exp(-(*x)*(*x)))

UNARY(j0, -j1(*x))
UNARY(j1,  j0(*x) - j1(*x)/(*x))
UNARY(y0, -y1(*x))
UNARY(y1,  y0(*x) - y1(*x)/(*x))

TEMPL inline CLASS jn(int n, const CLASS& x)
{
    CLASS r(jn(n, *x));
    REG(T(n)*jn(n, *x)/(*x) - jn(n+1, *x));
    return r;
}
TEMPL inline CLASS yn(int n, const CLASS& x)
{
    CLASS r(yn(n, *x));
    REG(T(n)*yn(n, *x)/(*x) - yn(n+1, *x));
    return r;
}


// complex support:
//
/***
TEMPL inline COMPLX conj(const COMPLX& x)
{
    COMPLX r(conj(*x));
    FOR(i,N) r[i] = conj(x[i]);
    return r;
}
TEMPL inline CLASS real(const COMPLX& x)
{
    CLASS r(real(*x));
    FOR(i,N) r[i] = real(x[i]);
    return r;
}
TEMPL inline CLASS imag(const COMPLX& x)
{
    CLASS r(imag(*x));
    FOR(i,N) r[i] = imag(x[i]);
    return r;
}
TEMPL inline CLASS arg(const COMPLX& x)
{
    return atan2(imag(x), real(x));
}
TEMPL inline CLASS abs(const COMPLX& x)
{
    return hypot(real(x), imag(x));
}
TEMPL inline CLASS norm(const COMPLX& x)
{
    return real(x)*real(x) + imag(x)*imag(x);
}
***/

// non-differentiables deliberately omitted:
// ~ ! & | && || ++ -- sign floor ceil etc.


