/*
  CRF++ -- Yet Another CRF toolkit

  $Id: tagger.cpp,v 1.8 2005/10/30 03:54:15 taku Exp $;

  Copyright (C) 2005 Taku Kudo <taku@chasen.org>

  This is free software with ABSOLUTELY NO WARRANTY.

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.	See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307	USA
*/
#include <iostream>
#include <vector>
#include <iterator>
#include <cmath>
#include "common.h"
#include "timer.h"
#include "tagger.h"
#include "lbfgs.h"

#define PRINT                                                           \
    for (size_t i = 0; i < x.size(); ++i) {                             \
      std::copy ( x[i].begin(), x[i].end(), std::ostream_iterator<char *>(os, "\t") ); \
      os << feature_index->get_y(result[i]);                            \
      if (vlevel >= 1) os << '/' << toprob(node[i][result[i]], Z);      \
      if (vlevel >= 2) {                                                \
        for (unsigned int j = 0; j < ysize; ++j)                        \
          os << '\t' << feature_index->get_y(j) << '/' << toprob(node[i][j], Z); \
      }                                                                 \
      os << '\n';                                                       \
    }                                                                   \

namespace CRFPP {

  static inline Fad<double> toprob (Node *n, Fad<double> Z)
  {
    return exp (n->alpha + n->beta - n->cost - Z);
    
  }

  bool Tagger::open (FeatureIndex *f)
  {
    mode = LEARN;
    feature_index = f;
    ysize = feature_index->ysize();
    return true;
  }

  bool Tagger::open (const char* filename)
  {
    mode = TEST;
    close ();
    feature_index = new DecoderFeatureIndex();
    if (! feature_index->open (filename, 0))
      WHAT_CLOSE_FALSE(feature_index->what());
    ysize = feature_index->ysize();
    return true;
  }

  bool Tagger::close ()
  {
    if (mode == TEST) {
      delete feature_index;
      feature_index = 0;
    }
    return true;
  }

  bool Tagger::parse (size_t nbest,
                      int vlevel,
                      std::istream &is,
                      std::ostream &os)
  {
    
    mode = TEST;

    if (! read (is)) return false;

    if (x.empty()) return true;

    buildLattice ();

    if (nbest || vlevel >= 1) forwardbackward ();

    viterbi ();

    os << std::endl;

    if (nbest >= 1) {
      initNbest ();
      for (size_t n = 0; n < nbest; ++n) {
        Fad<double> cost = 0.0;
        if (! next (cost)) break;
        os << "# " << n << " " << exp(-cost-Z) << std::endl;
        PRINT;
      }
    } else {
      if (vlevel >= 1) {
        Fad<double> cost = -node[x.size()-1][result[x.size()-1]]->bestCost;
        os << "# " << exp(-cost-Z) << std::endl;
      }
      PRINT;
    }

#undef PRINT

    return true;
  }

  bool Tagger::initNbest ()
  {
    nbest_freelist.free();
    while (! agenda.empty()) agenda.pop ();   // make empty

    size_t k = x.size()-1;
    for (size_t i = 0; i < ysize; ++i) {
      QueueElement *eos = nbest_freelist.alloc ();
      eos->node = node[k][i];
      eos->fx = -node[k][i]->bestCost;
      eos->gx = -node[k][i]->cost;
      eos->next = 0;
      agenda.push (eos);
    }

    return true;
  }

  bool Tagger::next (Fad<double> &score)
  {
    while (! agenda.empty()) {

      QueueElement *top = agenda.top ();
      agenda.pop ();
      Node *rnode = top->node;

      if (rnode->x == 0) {
        for (QueueElement *n = top; n; n = n->next)
          result[n->node->x] = n->node->y;
        score = top->gx;
        return true;
      }

      for (std::vector <Path *>::iterator it = rnode->lpath.begin(); it != rnode->lpath.end(); ++it) {
        QueueElement *n =nbest_freelist.alloc ();
        n->node = (*it)->lnode;
        n->gx	= - (*it)->lnode->cost	   - (*it)->cost +  top->gx;
        n->fx	= - (*it)->lnode->bestCost - (*it)->cost +  top->gx;
        //	    |		   h(x)			|  |  g(x)  |
        n->next = top;
        agenda.push (n);
      }
    }

    return 0;
  }

  void Tagger::eval (size_t &err,
                     size_t &all,
                     size_t &zeroone)
  {
    all += x.size();
    bool flg = false;
    for (size_t i = 0; i < x.size(); ++i)
      if (answer[i] != result[i]) { ++err; flg = true; }
    if (flg) ++zeroone;
  }

  void Tagger::clear ()
  {
    x.clear ();
    node.clear ();
    answer.clear ();
    result.clear ();
  }

  bool Tagger::empty()
  {
    return x.empty ();
  }

  bool Tagger::read (std::istream &is)
  {
    char  line[8192 * 16];
    char* column[256];
    clear ();

    size_t xsize = feature_index->xsize ();

    for (;;) {

      if (! is.getline (line, sizeof(line))) {
        is.clear (std::ios::eofbit|std::ios::badbit);
        return true;
      }

      if (line[0] == '\0' || line[0] == ' ' || line[0] == '\t') break;

      char *p = feature_index->strdup (line);

      size_t _size = tokenize (p, "\t ", column, 256);

      // std::cout<< *p << "  " << _size << "  " << *column << std::endl;
      
      if ((mode == LEARN && _size < xsize + 1) ||
          (mode == TEST	 && _size < xsize))
        WHAT_FALSE("# x is small");

      size_t s = x.size () + 1;
      x.resize	    (s);
      node.resize   (s);
      answer.resize (s);
      result.resize (s);
      s = x.size () - 1;

      for (size_t i = 0; i < _size; ++i){
        x[s].push_back (column[i]);
        // std::cout<< column[i] << std::endl;
      }
      answer[s] = 0; // dummy

      if (mode == LEARN) {
        size_t r = ysize;
        for (size_t k = 0; k < ysize; ++k)
          if (strcmp (feature_index->get_y(k), column[xsize]) == 0) r = k;

        if (r == ysize)
          WHAT_FALSE("cannot find answer: " << column[xsize]);

        answer[s] = r;
      }

      node[s].resize (ysize);
    }

    if (! feature_index->buildFeatures (x, node))
      WHAT_FALSE(feature_index->what());

    std::vector <std::vector <char *> >(x).swap (x);
    std::vector <std::vector <Node *> >(node).swap (node);
    std::vector <unsigned short int>(answer).swap (answer);
    std::vector <unsigned short int>(result).swap (result);

    return true;
  }

  void Tagger::buildLattice ()
  {
    if (x.empty()) return;

    feature_index->rebuildFeatures (x, node);

    for (size_t i = 0; i < x.size(); ++i) {
      for (size_t j = 0; j < ysize; ++j) {
                
        feature_index->calcCost (node[i][j]);
        std::vector <Path *> &lpath = node[i][j]->lpath;
        
        for (Path_iterator it = lpath.begin(); it != lpath.end(); ++it){
                    
          feature_index->calcCost (*it);
          
        }
      
      }
    
    }
  
  }

  void Tagger::forwardbackward ()
  {
    if (x.empty()) return;

    for (size_t i = 0; i < x.size(); ++i){
      
      for (size_t j = 0; j < ysize; ++j){
        
        node[i][j]->calcAlpha ();
        
      }
      
    }
    
    for (int i = (int)x.size()-1; i >= 0;  --i){
      
      for (size_t j = 0; j < ysize; ++j){
        
        // std::cout << i << "   " << j << std::endl;
        node[i][j]->calcBeta();
        // std::cout<< "bi: " << i << " j : " << j << "  " << node[i][j]-> beta << std::endl;
      }
      
    }
    
    Z = 0.0;
    for (size_t j = 0; j < ysize; ++j){
      // std::cout<< "Z before : " << Z << std::endl;
      // std::cout<< "beta : " << node[0][j]->beta << std::endl;
      Z = logsumexp (Z, node[0][j]->beta, j == 0);
      // std::cout<< "Z : " << Z << "  j:" << j << std::endl;
    }
    return;
  }

  void Tagger::viterbi ()
  {
    for (size_t i = 0;	 i < x.size(); ++i) {
      for (size_t j = 0; j < ysize; ++j) {
        Fad<double> bestc = -1e100;
        Node *best = 0;
        std::vector <Path *> &lpath = node[i][j]->lpath;
        for (Path_iterator it = lpath.begin(); it != lpath.end(); ++it) {
          Fad<double> cost = (*it)->lnode->bestCost + (*it)->cost + node[i][j]->cost;
          if (cost > bestc) {
            bestc = cost;
            best  = (*it)->lnode;
          }
        }
        
        node[i][j]->prev     = best;
        node[i][j]->bestCost = best ? bestc : node[i][j]->cost;
      
      }
    }

    Fad<double> bestc = -1e100;
    Node *best = 0;
    size_t s = x.size()-1;
    for (size_t j = 0; j < ysize; ++j) {
      if (bestc < node[s][j]->bestCost) {
        best  = node[s][j];
        bestc = node[s][j]->bestCost;
      }
    }

    for (Node *node = best; node; node = node->prev)
      result[node->x] = node->y;
    
  }

  double Tagger::calcGradient (Fad<double> *expected)
  {
    if (x.empty()) return 0.0;

    buildLattice ();

    forwardbackward ();

    double s = 0.0;
    
    for (size_t i = 0;	 i < x.size();	++i){
      
      for (size_t j = 0; j < ysize; ++j){
        
        node[i][j]->calcExpectation (expected, Z, ysize);
        
      }
    
    }
    
    // svnvish: I think he is just subtracting \phi(x_{i}, y_{i}) from
    // this point onwards
    // y =  answer
    // cost = phi(x_{i}, y_{i}) * theta
    
    for (size_t i = 0;	 i < x.size();	++i) {
      
      for (int *f = node[i][answer[i]]->fvector; *f != -1; ++f){
      
        --expected[*f + answer[i]];
        
      }
      
      s += node[i][answer[i]]->cost.v(); // UNIGRAM cost
      
      std::vector <Path *> &lpath = node[i][answer[i]]->lpath;
      
      for (Path_iterator it = lpath.begin(); it != lpath.end(); ++it) {
        
        if ((*it)->lnode->y == answer[(*it)->lnode->x]) {
        
          for (int *f = (*it)->fvector; *f != -1; ++f){
          
            --expected[*f + (*it)->lnode->y * ysize + (*it)->rnode->y];
            
          }
          
          s += (*it)->cost.v(); // BIGRAM COST
          
          break;
          
        }
        
      }
      
    }

    viterbi (); // call for eval ()
    
    // std::cout<< "Z:  " << Z << "  s :" << s << std::endl;
    
    return Z.v() - s ;
  
  }
  
  // calculate the update for the Collins perceptron algorithm
  double Tagger::calcCollins(Fad<double> *collins)
  {
    if (x.empty()) return 0.0;

    buildLattice ();

    forwardbackward ();

    viterbi (); // call for finding argmax y*

    double s = 0.0;

    // compute \phi(x_{i}, y*) - \phi(x_{i}, y_{i})
    // y* is the argmax computed via the viterbi call

    // y* = result[i]    
    // y = answer[i]

    for (size_t i = 0;	 i < x.size();	++i) {
      
      for (int *f = node[i][answer[i]]->fvector; *f != -1; ++f){
        
        --collins[*f + answer[i]];
        
      }

      s += node[i][answer[i]]->cost.v(); // UNIGRAM cost
      
      for (int *f = node[i][result[i]]->fvector; *f != -1; ++f){
        
        ++collins[*f + result[i]];
        
      }
            
      std::vector <Path *> &lpath = node[i][answer[i]]->lpath;
      
      for (Path_iterator it = lpath.begin(); it != lpath.end(); ++it) {
        
        if ((*it)->lnode->y == answer[(*it)->lnode->x]) {
        
          for (int *f = (*it)->fvector; *f != -1; ++f){
            
            --collins[*f + (*it)->lnode->y * ysize + (*it)->rnode->y];
            
          }
          s += (*it)->cost.v(); // BIGRAM COST 
          break;
        }
      }

      lpath = node[i][result[i]]->lpath;
      
      for (Path_iterator it = lpath.begin(); it != lpath.end(); ++it) {
        
        if ((*it)->lnode->y == result[(*it)->lnode->x]) {
          
          for (int *f = (*it)->fvector; *f != -1; ++f){
            
            ++collins[*f + (*it)->lnode->y * ysize + (*it)->rnode->y];
            
          }
          break;
        }
      }
    }
  
    return Z.v() - s ;
  }
  
}; 
